/*

	Color MDEF
	by Jordan Zimmerman
	(c)1995 All rights reserved worldwide.

	Unlimited use is hereby granted without restriction.  However,
	the author would appreciate credit if possible.  Please send
	comments, questions, bugs, etc. to jordanz@altura.com
	
	This code implements a Macintosh MDEF for selecting a color.
	The user is presented with a matrix of color chips to select
	from.
	
	Each menu item represents a color to be displayed. It should be
	a 12 byte string of the form:
		RRRRGGGGBBBB
	where RRRR, GGGG, and BBBB are hex numbers (as ASCII text).
	
	So, the color black would be:
		"\pFFFFFFFFFFFF"
		
		the color green would be:
		"\p0000FFFF0000"
		
	The current color is, by definition, the menu item that has a
	mark. So, you should only set a mark (any value) for one item
	at time.

	Change History:
	
	1.0.0	10/05/95	JLZ		Created inital version

*/

#include "color_mdef_utils.h"

const short	SQUARE_SIZE		= 13;
const short	SPACE_BETWEEN	= 2;
const short SELECTION_SIZE	= SPACE_BETWEEN;

const short	MENU_INSET		= SPACE_BETWEEN;

const short	POPUP_INSET		= 4;

// prototypes
static void draw_selection(short selected_square, int on_off_flag, short h_square_qty, short left, short top);
static void size_message(MenuHandle menu_h);
static void draw_message(MenuHandle menu_h, Rect& menu_r);
static void choose_message(MenuHandle menu_h, Rect& menu_r, Point hit_pt, short& which_item);
static void popup_message(MenuHandle menu_h, Rect& menu_r, short left, short top);
static short square_root(short number);
static void draw_color_square(const RGBColor color, short square_number, short h_square_qty, short left, short top);
static void get_square_rect(Rect& square_r, short square_number, short h_square_qty, short left, short top);
static short get_current_item(MenuHandle menu_h, short qty);

// MDEF Code Resource entry point
pascal void main(short message, MenuHandle menu_h, Rect* menu_r_ptr, Point hit_pt, short* which_item_ptr)
{

	char*		copyright_cstr = "Color MDEF v1.0 - by Jordan Zimmerman - All rights reserved worldwide.";

	switch ( message )
	{
		case mSizeMsg:
		{
			::size_message(menu_h);
			break;
		}
		
		case mDrawMsg:
		{
			::draw_message(menu_h, *menu_r_ptr);
			break;
		}	
		
		case mChooseMsg:
		{
			::choose_message(menu_h, *menu_r_ptr, hit_pt, *which_item_ptr);
			break;
		}
		
		case mPopUpMsg:
		{
			::popup_message(menu_h, *menu_r_ptr, hit_pt.v, hit_pt.h);	// yes, the MacOS reverses the meaning of hit_pit.h/v
			break;
		}
	}

} // main

// determine the size of the menu
// width/height is the square-root of the number of items times the size of a sqaure
static void size_message(MenuHandle menu_h)
{

	short		width = 0;
	short		height = 0;

	short		qty = ::CountMItems(menu_h);

	short		h_qty = ::square_root(qty);
	width = (h_qty * SQUARE_SIZE) + ((h_qty - 1) * SPACE_BETWEEN);

	short		v_qty = qty / h_qty;
	if ( qty % h_qty )
	{
		++v_qty;
	}
	height = (v_qty * SQUARE_SIZE) + ((v_qty - 1) * SPACE_BETWEEN);
	
	MenuInfo*	menu_ptr = *menu_h;
	menu_ptr->menuWidth = width + (MENU_INSET * 2);
	menu_ptr->menuHeight = height + (MENU_INSET * 2);

} // size_message

// draw all of the color chips - this only needs to be done once...
//... the selections don't affect it
static void draw_message(MenuHandle menu_h, Rect& menu_r)
{

	short		qty = ::CountMItems(menu_h);
	short		h_square_qty = ::square_root(qty);

	short		left = menu_r.left + MENU_INSET;
	short		top = menu_r.top + MENU_INSET;

	for ( short i = 0; i < qty; ++i )
	{
		Str255		pstr;
		::GetMenuItemText(menu_h, i + 1, pstr);
		
		RGBColor	color;
		::CMPstrToColor(pstr, color);
		::draw_color_square(color, i, h_square_qty, left, top);
		
		short		mark;
		::GetItemMark(menu_h, i + 1, &mark);
		if ( mark )
		{
			::draw_selection(i, true, h_square_qty, left, top);
		}
	}

} // draw_message

// do the hit testing
static void choose_message(MenuHandle menu_h, Rect& menu_r, Point hit_pt, short& which_item)
{

	short		qty = ::CountMItems(menu_h);
	
	// we need to remember the current item so that it "springs"
	// back to the current item if nothing is selected
	short		current_item = ::get_current_item(menu_h, qty);
	if ( !which_item )
	{
		which_item = current_item;
	}

	short		h_square_qty = ::square_root(qty);

	short		left = menu_r.left + MENU_INSET;
	short		top = menu_r.top + MENU_INSET;

	short new_which_item = qty;
	while ( new_which_item )
	{
		Rect	square_r;
		::get_square_rect(square_r, new_which_item - 1, h_square_qty, left, top);
		::InsetRect(&square_r, -SPACE_BETWEEN, -SPACE_BETWEEN);
		if ( ::PtInRect(hit_pt, &square_r) )
		{
			break;
		}
		
		--new_which_item;
	}
	if ( !new_which_item )
	{
		new_which_item = current_item;
	}
	
	if ( new_which_item != which_item )
	{	
		if ( which_item )
		{
			::draw_selection(which_item - 1, false, h_square_qty, left, top);
		}
		if ( new_which_item )
		{
			::draw_selection(new_which_item - 1, true, h_square_qty, left, top);
		}
		which_item = new_which_item;
	}

} // choose_message

// figure out the popup menu rectangle
static void popup_message(MenuHandle menu_h, Rect& menu_r, short left, short top)
{

	::CalcMenuSize(menu_h);	// this is what CDEF 0 does
	
	MenuInfo*	menu_ptr = *menu_h;
	menu_r.left = left;
	menu_r.top = top;
	menu_r.right = menu_r.left + menu_ptr->menuWidth;
	menu_r.bottom = menu_r.top + menu_ptr->menuHeight;
	
	GrafPtr		current_port_ptr;
	::GetPort(&current_port_ptr);
	
	// make sure it's on screen
	
	Rect		port_r = current_port_ptr->portRect;
	port_r.top += ::GetMBarHeight();
	::InsetRect(&port_r, POPUP_INSET, POPUP_INSET);
	
	short		h_diff = 0;
	if ( menu_r.left < port_r.left )
	{
		h_diff = port_r.left - menu_r.left;
	}
	else if ( menu_r.right > port_r.right )
	{
		h_diff = port_r.right - menu_r.right;
	}

	short		v_diff = 0;
	if ( menu_r.top < port_r.top )
	{
		v_diff = port_r.top - menu_r.top;
	}
	else if ( menu_r.bottom > port_r.bottom )
	{
		v_diff = port_r.bottom - menu_r.bottom;
	}
	
	// make sure it's not too big
	
	::OffsetRect(&menu_r, h_diff, v_diff);
	::SectRect(&menu_r, &port_r, &menu_r);

} // popup_message

// return the currently marked item
static short get_current_item(MenuHandle menu_h, short qty)
{

	// when the Menu Manager wants us to "flash" the new selection
	// it should keep going back and forth to the current item
	// So, recognize this by checking Button
	if ( !::Button() )
	{
		return 0;
	}

	for ( short i = 1; i <= qty; ++i )
	{
		short		mark;
		::GetItemMark(menu_h, i, &mark);
		if ( mark )
		{
			return i;
		}
	}
	
	return 0;

} // get_current_item

// draw one of the color chips
static void draw_color_square(const RGBColor color, short square_number, short h_square_qty, short left, short top)
{

	Rect		square_r;
	::get_square_rect(square_r, square_number, h_square_qty, left, top);

	::ForeColor(blackColor);
	::FrameRect(&square_r);

	::InsetRect(&square_r, 1, 1);
	
	::RGBForeColor(&color);
	::PaintRect(&square_r);

} // draw_color_square

// get the rectangle for any color chip
static void get_square_rect(Rect& square_r, short square_number, short h_square_qty, short left, short top)
{

	short		h_index = square_number % h_square_qty;
	short		v_index = square_number / h_square_qty;

	square_r.left = left + (h_index * (SQUARE_SIZE + SPACE_BETWEEN));
	square_r.top = top + (v_index * (SQUARE_SIZE + SPACE_BETWEEN));
	square_r.right = square_r.left + SQUARE_SIZE;
	square_r.bottom = square_r.top + SQUARE_SIZE;
	
} // get_square_rect

// draw the selection frame for any color chip
static void draw_selection(short selected_square, int on_off_flag, short h_square_qty, short left, short top)
{

	Rect		square_r;
	::get_square_rect(square_r, selected_square, h_square_qty, left, top);
	
	// for aestheics, change the chip's black frame to white
	// when selecting it - change it back to black when turning
	// off the selection
	::ForeColor(on_off_flag ? whiteColor : blackColor);
	::FrameRect(&square_r);

	::InsetRect(&square_r, -SELECTION_SIZE, -SELECTION_SIZE);
	::ForeColor(on_off_flag ? blackColor : whiteColor);
	::PenSize(SELECTION_SIZE, SELECTION_SIZE);
	::FrameRect(&square_r);
	
	::ForeColor(blackColor);
	::PenSize(1, 1);

} // draw_selection

// Brute-force square root
static short square_root(short number)
{

	if ( number > 256 )
	{
		return 16;	// error
	}
	
	if ( number < 4 )
	{
		return 2;
	}
	else if ( number <= 9 )
	{
		return 3;
	}
	else if ( number <= 16 )
	{
		return 4;
	}
	else if ( number <= 25 )
	{
		return 5;
	}
	else if ( number <= 36 )
	{
		return 6;
	}
	else if ( number <= 49 )
	{
		return 7;
	}
	else if ( number <= 64 )
	{
		return 8;
	}
	else if ( number <= 81 )
	{
		return 9;
	}
	else if ( number <= 100 )
	{
		return 10;
	}
	else if ( number <= 121 )
	{
		return 11;
	}
	else if ( number <= 144 )
	{
		return 12;
	}
	else if ( number <= 169 )
	{
		return 13;
	}
	else if ( number <= 196 )
	{
		return 14;
	}
	else if ( number <= 225 )
	{
		return 15;
	}
	else // if ( number <= 256 )
	{
		return 16;
	}

} // square_root
